/*
 * Copyright (c) 2012, Brandon Falk <bfalk@brandonfa.lk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 	1. Redistributions of source code must retain the above copyright notice,
 * 	this list of conditions and the following disclaimer.
 *
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ulpm.h"

UNICODE_STRING devsymname = RTL_CONSTANT_STRING(L"\\DosDevices\\ulpm");
PDEVICE_OBJECT devobj;

/* DriverEntry()
 *
 * Summary:
 *
 * This is the routine that is called upon loading of the driver. This is
 * comparable to the main() routine in user-mode programming.
 */
NTSTATUS
DriverEntry(PDRIVER_OBJECT drivobj, PUNICODE_STRING regpath)
{
	NTSTATUS       status  = STATUS_SUCCESS;
	UNICODE_STRING devname = RTL_CONSTANT_STRING(L"\\Device\\ulpm");
	int            i;

	UNREFERENCED_PARAMETER(regpath);

	dbgprint(">>> DriverEntry()\n");

	/* Set up the unload routine */
	drivobj->DriverUnload = drvunload;

	/* Set up our IRP callback routines */
	for(i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
		drivobj->MajorFunction[i] = genhdlr;

	drivobj->MajorFunction[IRP_MJ_DEVICE_CONTROL] = devctrl;
	drivobj->MajorFunction[IRP_MJ_CREATE]         = create;

	status = IoCreateDevice(
			drivobj,
			0,
			&devname,
			FILE_DEVICE_DATALINK,
			FILE_DEVICE_SECURE_OPEN,
			FALSE,
			&devobj);
	if(!NT_SUCCESS(status)){
		goto cleanup;
	}

	status = IoCreateSymbolicLink(&devsymname, &devname);
	if(!NT_SUCCESS(status)){
		IoDeleteDevice(devobj);
		goto cleanup;
	}

cleanup:
	dbgprint("<<< DriverEntry(0x%p)\n", status);

	return status;
}

/* drvunload()
 *
 * Summary:
 *
 * This function frees and unlocks any used resources. Called when the system
 * is going down, or when the driver is unloaded.
 */
VOID
drvunload(PDRIVER_OBJECT drivobj)
{
	UNREFERENCED_PARAMETER(drivobj);

	dbgprint(">>> drvunload()\n");

	IoDeleteSymbolicLink(&devsymname);
	IoDeleteDevice(devobj);

	dbgprint("<<< drvunload()\n");

	return;
}

/* genhdlr()
 *
 * Summary:
 *
 * This is a generic handler for unhandled IRPs.
 */
NTSTATUS
genhdlr(PDEVICE_OBJECT devobj, PIRP irp)
{
	PIO_STACK_LOCATION irpsp;

	UNREFERENCED_PARAMETER(devobj);

	irpsp = IoGetCurrentIrpStackLocation(irp);
	dbgprint("::: genhdlr(%s)\n", IRP_LOOKUP(irpsp->MajorFunction));

	irp->IoStatus.Status = STATUS_NOT_IMPLEMENTED;
	IoCompleteRequest(irp, IO_NO_INCREMENT);
	return STATUS_NOT_IMPLEMENTED;
}

/* create()
 *
 * Summary:
 *
 * This handles IRP_MJ_CREATE requests by returning STATUS_SUCCESS
 */
NTSTATUS
create(PDEVICE_OBJECT devobj, PIRP irp)
{
	UNREFERENCED_PARAMETER(devobj);
	UNREFERENCED_PARAMETER(irp);

	irp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest(irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}

/* devctrl()
 *
 * Summary:
 *
 * This function is the callback for IRP_MJ_DEVICE_CONTROL. Here we perform
 * the I/O to physical memory based on the request from user-mode.
 */
NTSTATUS
devctrl(PDEVICE_OBJECT devobj, PIRP irp)
{
	PIO_STACK_LOCATION  irpsp;
	struct _reqmem     *req;
	unsigned char      *ptr, *rawmem;
	unsigned __int32    i;
	int                 rs;
	PHYSICAL_ADDRESS    rawptr;

	UNREFERENCED_PARAMETER(devobj);

	irpsp = IoGetCurrentIrpStackLocation(irp);
	
	/* Filter out unknown requests */
	if(irpsp->Parameters.DeviceIoControl.IoControlCode != IOCTL_REQ_MEM){
		irp->IoStatus.Status = STATUS_NOT_IMPLEMENTED;
		IoCompleteRequest(irp, IO_NO_INCREMENT);
		return STATUS_NOT_IMPLEMENTED;
	}
	/* At this point, we have our proper request */

	/* Check if the input is the right size */
	if(irpsp->Parameters.DeviceIoControl.InputBufferLength <
			sizeof(struct _reqmem)){

		/* Send an error to user-mode */
		irp->IoStatus.Status      = STATUS_SUCCESS;
		irp->IoStatus.Information = PMERR_INPUT;
		IoCompleteRequest(irp, IO_NO_INCREMENT);
		return STATUS_SUCCESS;
	}

	req = irp->AssociatedIrp.SystemBuffer;

	/* Make sure the size does not exceed the buffer */
	if(req->size > irpsp->Parameters.DeviceIoControl.OutputBufferLength){

		/* Send an error to user-mode */
		irp->IoStatus.Status      = STATUS_SUCCESS;
		irp->IoStatus.Information = PMERR_OUTPUT;
		IoCompleteRequest(irp, IO_NO_INCREMENT);
		return STATUS_SUCCESS;
	}

	/* Check if we're 64-bit or not, then bounds check */
	if(Mm64BitPhysicalAddress){
		/* If the location is larger than our max location given the size
		 * requested
		 */
		if((0xFFFFFFFFFFFFFFFF - (req->size - 1)) < req->loc){
			/* Send an error to user-mode */
			irp->IoStatus.Status      = STATUS_SUCCESS;
			irp->IoStatus.Information = PMERR_BOUNDS;
			IoCompleteRequest(irp, IO_NO_INCREMENT);
			return STATUS_SUCCESS;
		}
	} else {
		/* If the location is larger than our max location given the size
		 * requested
		 */
		if(req->loc > 0xFFFFFFFF || (0xFFFFFFFF - (req->size - 1)) < req->loc){
			/* Send an error to user-mode */
			irp->IoStatus.Status      = STATUS_SUCCESS;
			irp->IoStatus.Information = PMERR_BOUNDS;
			IoCompleteRequest(irp, IO_NO_INCREMENT);
			return STATUS_SUCCESS;
		}
	}

	ptr = irp->UserBuffer;
	i   = req->size;
	rawptr.QuadPart = req->loc;

	/* Read the memory into user-mode in 64kB chunks */
	while(i){
		rs = (int)((i > (64 * 1024)) ? (64 * 1024) : i);

		rawmem = MmMapIoSpace(rawptr, rs, MmNonCached);
		if(!rawmem){
			/* Send an error to user-mode */
			irp->IoStatus.Status      = STATUS_SUCCESS;
			irp->IoStatus.Information = PMERR_INTERNAL;
			IoCompleteRequest(irp, IO_NO_INCREMENT);
			return STATUS_SUCCESS;
		}

		/* Do the I/O */
		if(!req->write)
			RtlCopyMemory(ptr, rawmem, rs);
		else
			RtlCopyMemory(rawmem, ptr, rs);

		MmUnmapIoSpace(rawmem, rs);

		rawptr.QuadPart += rs;
		i   -= rs;
		ptr += rs;
	}

	irp->IoStatus.Status      = STATUS_SUCCESS;
	irp->IoStatus.Information = 0;
	IoCompleteRequest(irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}

